/* 
   BlueMP3 firmware (c) 2004 by Till Harbaum, harbaum@beecon.de
*/

/*
   Dieser Code ist Teil der BlueMP3-Firmware wie den den Ausgaben
   8, 9 und 10 2004 des Computermagazins c't.

   Der Autor bernimmt keinerlei Garantie fr diesen Code. Dieser
   Code darf ohne Zustimmung des Autors in keiner Form weitervertrieben 
   werden.

   Dieser Code ist fr rein diadaktische Zwecke gedacht und ist 
   mglicherweise nicht kompatibel mit den Bluetooth-Standards.
*/

/*
   This code is part of the BlueMP3 firmware as published in issues
   8, 9, and 10 2004 of the german computer magazine c't. 

   The code comes without warantee of any kind, either expressed or
   implied. This code may not be distributed in any form without
   permission by the author.

   The code is provided for educational purposes only and may not comply 
   to the bluetooth standards.
*/

#include <avr/io.h>

#include "types.h"
#include "debug.h"

#include "cpu.h"
#include "console.h"
#include "timer.h"

#include "uart.h"
#include "hci.h"
#include "hci_callback.h"
#include "l2cap.h"

#include "vs1001.h"

#include "linux_client/l2send.h"

#define DEFAULT_VOLUME 90   // 90%

/* interfaces to the buttons */
#define BUTTON_PORT  PORTC
#define BUTTON_DDR   DDRC
#define BUTTON_PIN   PINC
#define BATT_LOW     4
#define BUTTON_0     5
#define BUTTON_1     6
#define BUTTON_2     7

#define LED_PORT     PORTC
#define LED_DIR      DDRC
#define LED_BLUETOOTH  0       // bluetooth activity LED
#define LED_POWER      1       // generic power LED

/* led update frequencies */
#define  UPDATE_FREQ   (50)   // 50 Hz
#define  LED_FAST      (4)    // 4 Hz
#define  LED_SLOW      (1)    // 1 Hz

/* string to be reported on request by host */
u08_t firmware_string[] = "xBlueMP3 L2CAP V" VERSION;

bool_t connected = FALSE;

/* check keys and return their state via l2cap */
void l2cap_check_buttons(hci_l2con_t *l2con) {
  static u08_t last_button_state = 0;
  u08_t button_message[2] = { MP3_BUTTONS, 0 };

  /* determine state of buttons, a released button is */
  if(!(inp(BUTTON_PIN) & BV(BUTTON_0))) button_message[1] |= 1;
  if(!(inp(BUTTON_PIN) & BV(BUTTON_1))) button_message[1] |= 2;
  if(!(inp(BUTTON_PIN) & BV(BUTTON_2))) button_message[1] |= 4;
  if(!(inp(BUTTON_PIN) & BV(BATT_LOW))) button_message[1] |= 8;

  /* buttons have changed */
  if(button_message[1] != last_button_state) {
    l2cap_send(l2con, button_message, 2);    
    last_button_state = button_message[1];
  }
}

bool_t hci_event_callback(hci_event_t *event) {
  bool_t ret = FALSE;
  
  switch(event->type) {
      /* react on an incoming connection request */
    case L2CAP_EVT_ACKNOWLEDGE_CONNECTION:
      DEBUG_MAIN("l2cap requesting acknowledge");

      /* allow nothing but MP3_PSM */
      ret = (event->l2con->psm == MP3_PSM);
      break;

     /* react on successful connection */
    case L2CAP_EVT_NOTIFY_CONNECTION:
      if(event->l2con->psm == MP3_PSM) {
	connected = TRUE;

        DEBUG_MAIN("l2cap connected");

	vs1001_soft_reset();
	vs1001_set_volume(100-DEFAULT_VOLUME, 100-DEFAULT_VOLUME);
      }
      break;

    case L2CAP_EVT_NOTIFY_DISCONNECTION:
      if(event->l2con->psm == MP3_PSM) {
	connected = FALSE;

        DEBUG_MAIN("l2cap disconnected");
      }
      break;

    case L2CAP_EVT_NOTIFY_RECEIVE:
      if(event->l2con->psm == MP3_PSM) {
	u16_t len = L2LEN(event->l2con);
	u08_t vol_r, vol_l, rst;

	/* check if this is a control packet */
	switch(l2cap_receive_u08(event->l2con)) {
	  case MP3_DATA:
	    /* ask mp3 layer to fetch data itself */
	    vs1001_receive(event->l2con, len-1);
	    break;

	  case MP3_RESET:
	    rst = l2cap_receive_u08(event->l2con);
	    l2cap_receive(event->l2con, NULL, len-2);
	    
	    DEBUG_MAIN("received reset command");
	    if(rst) vs1001_hard_reset();
	    else    vs1001_soft_reset();
	    
	    vs1001_set_volume(100-DEFAULT_VOLUME, 100-DEFAULT_VOLUME);
	    break;
	    
	  case MP3_VOLUME:
	    vol_l = l2cap_receive_u08(event->l2con);
	    vol_r = l2cap_receive_u08(event->l2con);
	    l2cap_receive(event->l2con, NULL, len-3);
	    
	    DEBUG_MAIN("received volume command");
	    vs1001_set_volume(vol_l, vol_r);
	    break;
  
	  case MP3_FWREV:
	    firmware_string[0] = MP3_FWREV;
	    l2cap_send(event->l2con, firmware_string, sizeof(firmware_string));
	    break;
	    
	  default:
	    DEBUG_MAIN("unsupported command");
	    l2cap_receive(event->l2con, NULL, len-1);
	    break;
	}

	/* and return button state if it has changed */
	l2cap_check_buttons(event->l2con);
      }
      /* once this function returns, all data not taken is dropped */
      break;

    default:
      break;
  }

  return ret;
}

int main(void) {
  ticks_t t, last_tick, last_led;

  /* initialize LEDs */
  sbi(LED_DIR, LED_POWER); 
  cbi(LED_PORT, LED_POWER); 
  sbi(LED_DIR, LED_BLUETOOTH);
  sbi(LED_PORT, LED_BLUETOOTH); 

  /* set direction for buttons and enable pullups */
  cbi(BUTTON_DDR, BUTTON_0);  sbi(BUTTON_PORT, BUTTON_0);
  cbi(BUTTON_DDR, BUTTON_1);  sbi(BUTTON_PORT, BUTTON_1);
  cbi(BUTTON_DDR, BUTTON_2);  sbi(BUTTON_PORT, BUTTON_2);
  cbi(BUTTON_DDR, BATT_LOW);  sbi(BUTTON_PORT, BATT_LOW);

  console_init();

  /* say hello on uart */
  DEBUG_MAIN("\r\rBlueMP3 V" VERSION " - " __DATE__ " " __TIME__ );

  /* start the system timer */
  timer_init();

  /* init mp3 chip and halt on error */
  if(!vs1001_init()) {
    DEBUG_MAIN("vs1001 initialisation failed, system halted\n");
    for(;;);
  }

  /* set initial volume */
  vs1001_set_volume(100-DEFAULT_VOLUME, 100-DEFAULT_VOLUME);

  /* this board supports resetting the device */
  /* after that wait for CTS to become deasserted */
  /* this routine makes sure, that cts is not reasserted */
  /* within a few milliseconds. This happens on the csr2 */
  /* and does _not_ mean, that the chip will be accepting data */
  sbi(DDRD, 4);
  sbi(PORTD, 4);
  cpu_delay(CPU_MILLISEC(50));
  cbi(PORTD, 4);
  do {
    while(inp(UART1_CTS_PORT) & (1<<UART1_CTS_BIT));
    cpu_delay(CPU_MILLISEC(50));
  } while(inp(UART1_CTS_PORT) & (1<<UART1_CTS_BIT));

  hci_init();

  DEBUG_MAIN("bluetooth hw ok");

  last_led = last_tick = timer_get_ticks();

  while(1) {
    hci_process(HCI_PROCESS_IDLE);

    t = timer_get_ticks();
    if(t - last_tick >= (100/UPDATE_FREQ)) {
      sbi(LED_PORT, LED_POWER);
      last_tick = t;

      /* are we somehow connected? */
      if(connected) {
        /* flash 2 hz if connected */
        if((t - last_led) > (100/LED_FAST)) {
          last_led = t;
	  cbi(LED_PORT, LED_POWER);
        }
      } else {
        /* flash 1 hz if not connected */
        if((t - last_led) > (100/LED_SLOW)) {
          last_led = t;
	  cbi(LED_PORT, LED_POWER);
        }
      }
    }
  }
  
  return 0;
}
