/*
  hci.c
*/
/* 
   BlueMP3 firmware (c) 2004 by Till Harbaum, harbaum@beecon.de
*/

/*
   Dieser Code ist Teil der BlueMP3-Firmware wie den den Ausgaben
   8, 9 und 10 2004 des Computermagazins c't.

   Der Autor bernimmt keinerlei Garantie fr diesen Code. Dieser
   Code darf ohne Zustimmung des Autors in keiner Form weitervertrieben 
   werden.

   Dieser Code ist fr rein diadaktische Zwecke gedacht und ist 
   mglicherweise nicht kompatibel mit den Bluetooth-Standards.
*/

/*
   This code is part of the BlueMP3 firmware as published in issues
   8, 9, and 10 2004 of the german computer magazine c't. 

   The code comes without warantee of any kind, either expressed or
   implied. This code may not be distributed in any form without
   permission by the author.

   The code is provided for educational purposes only and may not comply 
   to the bluetooth standards.
*/

#include "types.h"
#include "debug.h"

#include "hci.h"
#include "hci_uart.h"
#include "hci_event.h"
#include "hci_info.h"
#include "hci_acl.h"
#include "hci_con.h"
#include "utils.h"
#include "uart.h"
#include "cpu.h"

#include <stdarg.h>

#define MK_U16(a)      ((a)&0xff), (((a)>>8)&0xff)
#define HCI_SEC16(a)   MK_U16((a)*1600)
#define HCI_MSEC16(a)  MK_U16((a)*1.6)

u08_t hci_device_name[] = DEVICE_NAME;

/* load hardware specific code */
#include "hci_csr.h"

static u08_t hci_decode(u16_t cmd, void *data) {
  u08_t result = HCI_DECODE_CONT;  /* command has not been ack'd yet */

  HCI_ACTIVITY_ON;

  /* fetch hci message type byte */
  switch(hci_uart_get()) {
    case HCI_ACL:
      hci_acl_decode(cmd);
      break;

    case HCI_EVT:
      result = hci_event(cmd, data);
      break;

    default:
      DEBUG_HCI("PANIC: Received invalid HCI UART message type");
      for(;;);
  }

  HCI_ACTIVITY_OFF;
  return result;
}

/* start the transmission of a hci command */
static void hci_cmd_start(void) {
  HCI_ACTIVITY_ON;

  /* wait for free command buffers */
  while(!hci_further_commands)
    hci_process(HCI_PROCESS_WAIT_CMD);

  hci_uart_put(HCI_CMD);   // this will be a hci command
}

#define GET_U16LE(a)  ((a)[0]+256u*(a)[1])

/* wait for the command to be processed */
static bool_t hci_wait(u08_t *cmd, void *data) {
  u08_t ret;
  u16_t timeout = 30000;

  /* wait for serial data (with timeout) */
  while((!uart_input_avail())&&(timeout))
    timeout--;

  if(!timeout) {
    DEBUG_HCI("command reply timeout");

    HCI_ACTIVITY_OFF;
    return FALSE;       // retry due to timeout
  }

  /* wait for command complete, this command blocks and if no */
  /* answer is received within the watch dog timeout, the device */
  /* is being reset. this is ok, since _all_ commands should */
  /* trigger a reply within few milliseconds. see hci_acl.c for */
  /* the only exception to this rule. */
  while((ret = hci_decode(GET_U16LE(cmd), data)) == HCI_DECODE_CONT);

  HCI_ACTIVITY_OFF;

  /* return FALSE on error (hardware out of sync) */
  return(ret == HCI_DECODE_OK);
}

/* send a byte sequence */
void hci_send_seq(u08_t *data, u08_t len) {
  if(data) hci_uart_send(data, len); 
  else     hci_uart_send_zero(len);
}

/* send a hci command consisting of one single byte sequence */
void hci_send_cmd(u08_t *cmd, u08_t len, u08_t par_num, ...) {
  va_list ap;
  u08_t *pptr, cnt;
  u16_t plen;
  void  *data = NULL;

  do {
    va_start(ap, par_num);

    /* send cmd */
    hci_cmd_start();
    hci_uart_send(cmd, len);

    /* is a result expected? */
    if(par_num & HCI_PARMS) {
      /* yes, get buffer address */
      data = va_arg(ap, void *);
      par_num &= ~HCI_PARMS;
    }

    /* send parameters */
    for(cnt=0;cnt<par_num;cnt++) {

      /* get address and length */
      pptr = va_arg(ap, u08_t *);
      plen = va_arg(ap, int);

      /* and send it from flash, eeprom, ram, ... */
      switch(plen & HCI_FLAG_MASK) {
	case HCI_RAM:
	  hci_uart_send(pptr, plen);  // no mask req., since HCI_RAM == 0
	  break;

	case HCI_ZERO:
	  hci_uart_send_zero(plen & HCI_LEN_MASK);
	  break;
      }
    }

    va_end(ap);

    /* wait for cmd to be acked, hci_wait returns FALSE on hardware */ 
    /* error. This indicates, that the hci layer is out of sync and */
    /* the last command is very likely lost */
  } while(!hci_wait(cmd, data));
}

/* set name from eeprom */
void hci_set_name(void) {
  static u08_t hci_cmd_chname[] = { 
    OPCODE(OGF_HOST_CTL, OCF_CHANGE_LOCAL_NAME), 248 };

  hci_send_cmd(hci_cmd_chname, sizeof(hci_cmd_chname), 2,
	       hci_device_name, sizeof(hci_device_name),
	       NULL,            HCI_ZERO | (248 - sizeof(hci_device_name)));
}

bool_t hci_reset(void) {
  u08_t to = 100;  // 100*10ms timeout
  static u08_t hci_cmd_reset[] = { 
    OPCODE(OGF_HOST_CTL, OCF_RESET), 0 };

  /* send reset cmd */
  hci_cmd_start();
  hci_uart_send(hci_cmd_reset, sizeof(hci_cmd_reset));

  /* and wait for some answer */
  while(!uart_input_avail()) {
    cpu_delay(CPU_MILLISEC(10));

    if(!to--) return FALSE;
  }

  /* there's some reply */
  return hci_wait(hci_cmd_reset, NULL);
}

/* send all commands necessary to set up the bluetooth interface */
void hci_init(void) {
  static u08_t hci_cmd_write_cod[] = { 
    OPCODE(OGF_HOST_CTL, OCF_WRITE_COD), 3, MK_SEQ3(DEVICE_CLASS) };

  static u08_t hci_cmd_inq_enable[] = {
    OPCODE(OGF_HOST_CTL, OCF_WRITE_SCAN_ENABLE), 1,
    PAGE_SCAN_ENABLE | INQUIRY_SCAN_ENABLE };

  DEBUG_HCI("hci_init()");

  /* setup/clear connection structures etc */
  hci_con_init();

  /* check whether device already runs on alternate baud rate */

  /* send reset command */
  if(!hci_reset()) {
    DEBUG_HCI("reset at default baudrate timed out");

    uart_set_bitrate();
    if(!hci_reset()) {    
      DEBUG_HCI("reset at alternate baudrate timed out, halting");
      for(;;);
    } 
  }

  DEBUG_HCI("bluetooth reset successful");

  /* use csr specific vendor commands to increase bitrate */
  hci_csr_init();

  /* set name of device */
  DEBUG_HCI("setting name");
  hci_set_name();

  /* set class of device */
  DEBUG_HCI("setting COD");
  hci_send_cmd(hci_cmd_write_cod, sizeof(hci_cmd_write_cod), 0);

  /* enable the device to be discovered */
  DEBUG_HCI("write page enable");
  hci_send_cmd(hci_cmd_inq_enable, sizeof(hci_cmd_inq_enable), 0);

  /* send initial init commands */
  hci_info_init();
}

void hci_process(u16_t cmd) {
  /* process data if byte on uart waiting or buffers in use */
  if(uart_input_avail()) {

    /* decode without a pending command */
    hci_decode(cmd, NULL);
  }
}
