/*
  timer.c

  timer routines for ATmega128, implements a 100hz system timer
*/
/* 
   BlueMP3 firmware (c) 2004 by Till Harbaum, harbaum@beecon.de
*/

/*
   Dieser Code ist Teil der BlueMP3-Firmware wie den den Ausgaben
   8, 9 und 10 2004 des Computermagazins c't.

   Der Autor bernimmt keinerlei Garantie fr diesen Code. Dieser
   Code darf ohne Zustimmung des Autors in keiner Form weitervertrieben 
   werden.

   Dieser Code ist fr rein diadaktische Zwecke gedacht und ist 
   mglicherweise nicht kompatibel mit den Bluetooth-Standards.
*/

/*
   This code is part of the BlueMP3 firmware as published in issues
   8, 9, and 10 2004 of the german computer magazine c't. 

   The code comes without warantee of any kind, either expressed or
   implied. This code may not be distributed in any form without
   permission by the author.

   The code is provided for educational purposes only and may not comply 
   to the bluetooth standards.
*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/signal.h>

#include "types.h"
#include "debug.h"
#include "timer.h"
#include "cpu.h"

volatile ticks_t timer_ticks;

#if ((CPU_CLOCK / 65536) <= TIMER_HZ )
// no prescaler required
#define TIMER_PRESCALE (1<<CS10)
#define TIMER_RELOAD   (u16_t)(((CPU_CLOCK/1) / TIMER_HZ)-1)
#define TIMER_REAL     (CPU_CLOCK/((TIMER_RELOAD+1)*1))
#elif (((CPU_CLOCK/8) / 65536) <= TIMER_HZ )
#define TIMER_PRESCALE (1<<CS11)
#define TIMER_RELOAD   (u16_t)(((CPU_CLOCK/8) / TIMER_HZ)-1)
#define TIMER_REAL     (CPU_CLOCK/(TIMER_RELOAD+1)/8)
#elif (((CPU_CLOCK/64) / 65536) <= TIMER_HZ )
#define TIMER_PRESCALE ((1<<CS11)|(1<<CS10))
#define TIMER_RELOAD   (u16_t)(((CPU_CLOCK/64) / TIMER_HZ)-1)
#define TIMER_REAL     (CPU_CLOCK/(TIMER_RELOAD+1)/64)
#else
// add more setups here (atmel datasheet pg 133)
#error No appropriate prescaler setting found!
#endif

SIGNAL(SIG_OUTPUT_COMPARE1A) {
  /* clear counter */
  TCNT1 = 0;  
  timer_ticks++;
}

/* init timer hardware and install interrupt handler */
void timer_init(void) {
  timer_ticks = 0;

  // make counter run
  TCCR1B = TIMER_PRESCALE; // timer1 prescaler CK/8 = 916kHz @ 7.328 Mhz

  /* 16 bit overflow = CPU_CLK/65536, at 916 kHz hz = 13.977 Hz */
  OCR1A = TIMER_RELOAD;

  // make counter generate interrupts
  TIMSK |= BV(OCIE1A); // timer 1 compare a event

  sei();
}


