/*
  uart.c

  simple avr atmega128 uart routines
*/
/* 
   BlueMP3 firmware (c) 2004 by Till Harbaum, harbaum@beecon.de
*/

/*
   Dieser Code ist Teil der BlueMP3-Firmware wie den den Ausgaben
   8, 9 und 10 2004 des Computermagazins c't.

   Der Autor bernimmt keinerlei Garantie fr diesen Code. Dieser
   Code darf ohne Zustimmung des Autors in keiner Form weitervertrieben 
   werden.

   Dieser Code ist fr rein diadaktische Zwecke gedacht und ist 
   mglicherweise nicht kompatibel mit den Bluetooth-Standards.
*/

/*
   This code is part of the BlueMP3 firmware as published in issues
   8, 9, and 10 2004 of the german computer magazine c't. 

   The code comes without warantee of any kind, either expressed or
   implied. This code may not be distributed in any form without
   permission by the author.

   The code is provided for educational purposes only and may not comply 
   to the bluetooth standards.
*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/signal.h>

#include "types.h"
#include "debug.h"

#include "console.h"
#include "cpu.h"
#include "uart.h"

u08_t uart_rx1_buf[256];
u08_t uart_rx1_rd, uart_rx1_wr;

void uart_init(void) {
  /* init rx buffer read and write pointer */
  uart_rx1_rd = uart_rx1_wr = 0;

  /* set uart 0 baudrate */
  UBRR0H = (u08_t)(DEFAULT_BAUDRATE_REG>>8);
  UBRR0L = (u08_t)(DEFAULT_BAUDRATE_REG);

  /* switch uart to 8 bits, no parity, 1 stop bit */
  UCSR0C = 3 << UCSZ0;
  /* Enable UART0 transmitter */
  UCSR0B = BV(TXEN);

  /* Set uart 1 baud rate */
  UBRR1H = (u08_t)(DEFAULT_BAUDRATE_REG>>8);
  UBRR1L = (u08_t)(DEFAULT_BAUDRATE_REG);
  
  /* Set frame format: 8 data 1 stop */
  UCSR1C = 3 << UCSZ0;

  sbi(UART1_RTS_DDR, UART1_RTS_BIT);       // set RTS to output
  cbi(UART1_RTS_PORT, UART1_RTS_BIT);      // enable reception

  /* Enable UART receiver and transmitter */
  UCSR1B = BV(RXEN)|BV(RXCIE)|BV(TXEN);

  /* enable interrupts */
  sei();
}

void  uart_set_bitrate(void) {
  cpu_delay(CPU_MILLISEC(30));

  DEBUG_HCI("changing HCI_UART baudrate");

  /* set the new baud rate */
  UBRR1H = (u08_t)(HCI_BAUDRATE_REG>>8);
  UBRR1L = (u08_t)(HCI_BAUDRATE_REG);
}

void  uart_enable_rx(bool_t on) {
  cpu_delay(CPU_MILLISEC(30));

  /* enable/disable rts on bluetooth uart */
  if(on) cbi(UART1_RTS_PORT, UART1_RTS_BIT);      // enable reception
  else   sbi(UART1_RTS_PORT, UART1_RTS_BIT);      // disable reception
}

SIGNAL(SIG_UART1_RECV) {
  u08_t offset;

  /* irq driven input, we don't have to care about pointer */
  /* wrapping since the buffer is 256 bytes in size and we use */
  /* 8 bit counters */
  uart_rx1_buf[uart_rx1_wr++] = UDR1;
  offset = uart_rx1_rd - uart_rx1_wr;

  /* is the buffer nearly full? */
  if(offset < UART1_RTS_OFFSET ) 
    sbi(UART1_RTS_PORT, UART1_RTS_BIT);      // disable reception
}

/* check if data is available from uart1 (bluetooth uart) */
u08_t uart_input_avail(void) {
  return (uart_rx1_rd - uart_rx1_wr);
}

/* read a byte from uart1 (bluetooth uart) */
u08_t uart_getc(void) {
  /* read value from buffer */
  u08_t ret = uart_rx1_buf[uart_rx1_rd++];
  u08_t offset = uart_rx1_rd - uart_rx1_wr;

  if(offset > UART1_RTS_OFFSET )
    cbi(UART1_RTS_PORT, UART1_RTS_BIT);      // enable reception

  return ret;
}

void uart_putc(u08_t uart, u08_t chr) {
  if(uart == UART0) {
    /* Wait for empty transmit buffer */
    while( !(UCSR0A & BV(UDRE)) );

    /* Start transmission */
    UDR0 = chr;
  } else {
    /* wait for CTS to become asserted */
    while(inp(UART1_CTS_PORT) & (1<<UART1_CTS_BIT));

    /* Wait for empty transmit buffer */
    while ( !(UCSR1A & BV(UDRE)) );

    /* Start transmission */
    UDR1 = chr;
  }
}

