/*
  vs1001.c

  vs1001 mp3 dsp driver routines
*/
/* 
   BlueMP3 firmware (c) 2004 by Till Harbaum, harbaum@beecon.de
*/

/*
   Dieser Code ist Teil der BlueMP3-Firmware wie den den Ausgaben
   8, 9 und 10 2004 des Computermagazins c't.

   Der Autor bernimmt keinerlei Garantie fr diesen Code. Dieser
   Code darf ohne Zustimmung des Autors in keiner Form weitervertrieben 
   werden.

   Dieser Code ist fr rein diadaktische Zwecke gedacht und ist 
   mglicherweise nicht kompatibel mit den Bluetooth-Standards.
*/

/*
   This code is part of the BlueMP3 firmware as published in issues
   8, 9, and 10 2004 of the german computer magazine c't. 

   The code comes without warantee of any kind, either expressed or
   implied. This code may not be distributed in any form without
   permission by the author.

   The code is provided for educational purposes only and may not comply 
   to the bluetooth standards.
*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/signal.h>

#include "types.h"
#include "debug.h"

#include "cpu.h"
#include "vs1001.h"

#include "l2cap.h"

#define CLOCK_UNKNOWN   0
#define CLOCK_12MHZ     1
#define CLOCK_24MHZ     2

u08_t clock = CLOCK_UNKNOWN;

/* 2 kByte of buffer */
#define BUFFER_SIZE  (64*32)    // size must be multiple of 32

volatile static u08_t burst_count;     // no burst in progress
volatile static u08_t buffer[BUFFER_SIZE];
volatile static u08_t *r_ptr;
volatile static u16_t r_index, w_index;

volatile static bool_t force_restart;

/* send a single byte via spi */
static inline void spi_write_byte(u08_t byte) {
  outp(byte, SPDR);
  loop_until_bit_is_set(SPSR, SPIF);    
}

/* receive a single byte via spi */
static inline u08_t spi_read_byte(void) {
  outp(0, SPDR);
  loop_until_bit_is_set(SPSR, SPIF);    
  return SPDR;
}

static inline void vs1001_start_burst(void) {
  burst_count = 32;               // 32 byte burst in progress
  r_ptr = &buffer[r_index];       // use new chunk
  r_index += 32;
  if(r_index == BUFFER_SIZE) r_index = 0;

  sbi(VS1001_BSYNC_PORT, VS1001_BSYNC_BIT); 	   // byte sync active
  outp(*r_ptr++, SPDR);           // write first byte of 32 bytes sequence
}

static bool_t vs1001_restart(void) {
  u16_t avail;

  /* check if enough bytes are available */
  if(r_index <= w_index) avail = w_index - r_index;
  else                   avail = w_index + (BUFFER_SIZE - r_index);
  
  if(avail >= 32) {
    vs1001_start_burst();
    return TRUE;
  }
  return FALSE;
}

  
/* dreq has gone high, dsp is requesting more data */
/* if no 32 byte burst is in progress, start a new */
/* one from the second buffer */
SIGNAL(VS1001_DREQ_SIG) {

  /* try to restart here if spi wasn't able to restart itself */
  /* request mainloop to start if it doesn't work here */
  /* due to no data being available */
  if(!burst_count)
    if(!vs1001_restart())
      force_restart = TRUE;
}

/**************************************************************
 **************************************************************/

SIGNAL(SIG_SPI) {
  if(burst_count > 1) {
    /* we are still in a burst */
    burst_count--;                   // decrease burst counter
    outp(*r_ptr++, SPDR);            // write next byte
  } else {
    if(burst_count) {
      /* burst ended, deactivate byte sync */
      cbi(VS1001_BSYNC_PORT, VS1001_BSYNC_BIT);

      /* check if the transfer might be restarted instantly */
      /* if successful, burst_count will be reloaded from 1 to 32 */

      /* if dreq is active -> try to restart transfer */
      if(inp(VS1001_DREQ_PORT) & (1<<VS1001_DREQ_BIT)) {
	if(!vs1001_restart()) {
	  /* unable to restart -> buffer is empty */
	  /* ask main loop to restart once new data */
	  /* has been put into the buffer */
	  force_restart = TRUE;    
	  burst_count = 0;
	}
      } else
	burst_count = 0;   // else wait for dreq irq to restart
    }
  }
}

/* reset the VS1001 */
void vs1001_reset(void) {
  cbi(VS1001_RESET_PORT, VS1001_RESET_BIT);	// reset active
  cpu_delay(CPU_MILLISEC(1));                   // wait 1 msec	    
  sbi(VS1001_RESET_PORT, VS1001_RESET_BIT);	// reset inactive
  cpu_delay(CPU_MILLISEC(2));                   // wait 2 msec	    
}

/* read a word from configuration inferface */
u16_t vs1001_read(u08_t addr) {
  u16_t retval;

  cbi(VS1001_XCS_PORT, VS1001_XCS_BIT);	// xCS lo

  spi_write_byte(VS1001_READ);
  spi_write_byte(addr);
  retval  = ((u16_t)spi_read_byte()) << 8;
  retval |= spi_read_byte();
    
  sbi(VS1001_XCS_PORT, VS1001_XCS_BIT);	// xCS hi
  
  // spec wants 5 microseconds delay after spc io
  cpu_delay(CPU_MICROSEC(5)); 

  return retval;
}

/* write a word to configuration inferface */
void vs1001_write(u08_t addr, u16_t value) {

  cbi( VS1001_XCS_PORT, VS1001_XCS_BIT);	// xCS lo

  spi_write_byte(VS1001_WRITE);
  spi_write_byte(addr);
  spi_write_byte(value>>8);
  spi_write_byte(value);

  sbi(VS1001_XCS_PORT, VS1001_XCS_BIT);	       // xCS hi
  
  // spec wants 5 microseconds delay after spc io
  cpu_delay(CPU_MICROSEC(5)); 
}

void vs1001_hard_reset(void) {
  DEBUG_MP3("performing DSP hard reset");

  cbi(SPCR, SPIE);        // SPI interrupt disable
  cbi(EIMSK, INT4);       // disable dreq irq

  /* start system for normal operation */
  vs1001_reset();
  
  /* send zero command */
  vs1001_zero(8);

  if(clock == CLOCK_12MHZ) {
    DEBUG_MP3("detected low dsp clock");
    vs1001_write(VS1001_CLOCKF, 0x8000 + VS1001_CLOCK_12MHZ/2000);  
  } else {
    DEBUG_MP3("detected high dsp clock");
    vs1001_write(VS1001_CLOCKF, 0);
  }

  /* init buffer pointers */
  w_index = r_index = 0;

  /* restart from main loop */
  force_restart = TRUE;

  /* no burst in progress */
  burst_count = 0;

  sbi(SPCR, SPIE);        // SPI interrupt enable
  sbi(EIMSK, INT4);       // enable dreq irq
}

/* issue a soft reset, the datasheet wants this to be done every new song */
void vs1001_soft_reset(void) {
  u16_t timeout;

  DEBUG_MP3("performing DSP soft reset");

  /* wait for for dsp running out of data  */
  /* give DSP at most 2 seconds */
  timeout = 2000;
  while((!force_restart)&&(timeout)) {
    cpu_delay(CPU_MILLISEC(1));
    timeout--;
  }

  cbi(SPCR, SPIE);        // SPI interrupt disable
  cbi(EIMSK, INT4);       // disable dreq irq

  /* send reset command */
  vs1001_write(VS1001_MODE, VS1001_SM_RESET);
  cpu_delay(CPU_MICROSEC(2));

  /* wait for DREQ to become asserted, hard reset if this fails */
  /* within 1 msec */
  timeout = 1000;
  while((inp(VS1001_DREQ_PORT) & (1<<VS1001_DREQ_BIT))&&(timeout)) {
    cpu_delay(CPU_MICROSEC(1));
    timeout--;
  }

  if(!timeout) {    
    DEBUG_MP3("soft reset timeout, doing hard reset");
    vs1001_hard_reset();
  } else {
    /* init buffer pointers */
    w_index = r_index = 0;

    /* send some zeros */
    vs1001_zero(8);
    
    /* restart from main loop */
    force_restart = TRUE;
    
    /* no burst in progress */
    burst_count = 0;
    
    sbi(SPCR, SPIE);        // SPI interrupt enable
    sbi(EIMSK, INT4);       // enable dreq irq
  }
}

/* send a special memory test sequence to the VS1001 MPEG */
/* stream interface */
void vs1001_send_test(void) {
  DEBUG_MP3("sending test command sequence");

  /* according to spec, test sequence starts with resetting the hardware */
  vs1001_reset();

  sbi(VS1001_BSYNC_PORT, VS1001_BSYNC_BIT);        // byte sync active

  spi_write_byte(0x4d);                        // send four data bytes
  spi_write_byte(0xea);
  spi_write_byte(0x6d);
  spi_write_byte(0x54);
  spi_write_byte(0);		                   // send four zero bytes
  spi_write_byte(0);
  spi_write_byte(0);
  spi_write_byte(0);

  cbi(VS1001_BSYNC_PORT, VS1001_BSYNC_BIT);        // byte sync inactive
}

/* send a burst of 32 data bytes to the VS1001 MPEG stream, since this */
/* is the size of data chunks, DREQ requests for */
void vs1001_receive(hci_l2con_t *l2con, u16_t len) {
  u08_t *buf, bytes2copy;
  u16_t space;

  /* wait for buffer to become available */
  /* blocking here makes the bluetooth connection halt, which is */
  /* exactly what we want */

  /* read/receive data chunk by chunk */
  while(len) {
    /* check if there's enough buffer space for yet another chunk */
    do {
      /* determine space in buffer */
      if(w_index < r_index) space = r_index - w_index - 33;
      else                  space = r_index + (BUFFER_SIZE - w_index) - 33;    

      /* do _not_ reset watch dog timer during wait, since wating too long */
      /* means that the DSP stopped accepting data which is a major failure */
    } while(space < 32);

    /* ok, if we get here, buffer is available -> fill it! */
    /* buffer may be wrapped (first part has to be stored */
    /* at the end of the buffer, the remaining bytes at */
    /* its beginning) */
    
    buf = (u08_t*)(buffer + w_index);
    if(len >= 32) bytes2copy = 32;
    else          bytes2copy = len;

    len -= bytes2copy;

    if((w_index + bytes2copy) < BUFFER_SIZE) {
      /* write all 32 bytes at once */
      l2cap_receive(l2con, buf, bytes2copy);
      w_index += bytes2copy;
    } else {
      u16_t frag_len = BUFFER_SIZE-w_index;
      
      /* write first fragment */
      l2cap_receive(l2con, buf, frag_len);
      
      /* write second fragment */
      buf = (u08_t*)buffer;               // restart at buffer start
      frag_len = bytes2copy - frag_len;   // remaining bytes
      l2cap_receive(l2con, buf, frag_len);
      w_index = frag_len;
    }
    
    /* start frm mainloop if requested by spi and dsp really */
    /* wants data */
    if(force_restart && (inp(VS1001_DREQ_PORT) & (1<<VS1001_DREQ_BIT)))
      if(vs1001_restart())
	force_restart = FALSE;
  }
}

/* send zeros to the data interface (used internally without irqs) */
void vs1001_zero(u08_t num) {
  sbi(VS1001_BSYNC_PORT, VS1001_BSYNC_BIT);        // byte sync active
  while(num--) spi_write_byte(0);		   // send zero bytes
  cbi(VS1001_BSYNC_PORT, VS1001_BSYNC_BIT);        // byte sync inactive
}

/* ask vs1001 to perform memory test */
bool_t vs1001_memory_test(void) {
  DEBUG_MP3("doing vs1001k memory test ...");

  /* Send sequence and wait 500000 clock cycles according to spec */
  /* which is about 21 msec @ 24.5Mhz. My prototype required 60 msec */
  /* running on a 24Mhz quartz. This explained the fact, that i saw */
  /* about 8Mhz when watching the clock signal with a scope ... */
  vs1001_send_test();

  cpu_delay(CPU_MILLISEC(25)); 

  if(vs1001_read(VS1001_HDAT0) == 0x7f) {
    DEBUG_MP3("Successful after 25msec -> 24.5 Mhz");
    clock = CLOCK_24MHZ;
    return TRUE;
  }

  cpu_delay(CPU_MILLISEC(25)); 

  if(vs1001_read(VS1001_HDAT0) == 0x7f) {
    DEBUG_MP3("Successful after 50msec -> 12 Mhz");
    clock = CLOCK_12MHZ;
    return TRUE;
  }

  return FALSE;
}

// setup the VS1001 chip for decoding
bool_t vs1001_init(void) {

  // DREQ is input
  cbi(VS1001_DREQ_DDR,   VS1001_DREQ_BIT);

  // BSYNC is output, default low
  sbi(VS1001_BSYNC_DDR,  VS1001_BSYNC_BIT);
  cbi(VS1001_BSYNC_PORT, VS1001_BSYNC_BIT);
  
  // chip select is output, default hi
  sbi(VS1001_XCS_DDR,    VS1001_XCS_BIT);
  sbi(VS1001_XCS_PORT,   VS1001_XCS_BIT);

  // reset is output, default low
  sbi(VS1001_RESET_DDR,  VS1001_RESET_BIT);
  sbi(VS1001_RESET_PORT, VS1001_RESET_BIT);

  /* initialize sp interface */
  sbi(DDRB, 0);       // SS -> output 
  sbi(DDRB, 1);	      // set SCK as output
  sbi(DDRB, 2);

  cbi(PORTB, 1);      // set SCK lo
	
  cbi(SPCR, SPIE);   // SPI interrupt disable
  cbi(SPCR, DORD);   // LSB first
  sbi(SPCR, MSTR);   // we are master
  cbi(SPCR, CPOL);   // data valid at falling edge
  cbi(SPCR, CPHA);   // data valid at edge's end
  sbi(SPCR, SPR0);   // use XTAL/4 as SPI clock

  sbi(SPCR, SPE);    // SPI enable


  /* setup DREQ interrupt */
  cbi(VS1001_DREQ_DDR, VS1001_DREQ_BIT);

  if(!vs1001_memory_test()) {
    DEBUG_MP3("memory test failed");
    return FALSE;
  }

  /* configure INT4 (dreq irq) */
  sbi(EICRB, ISC40);      // rising edge
  sbi(EICRB, ISC41);

  /* memory test ok, re-start system for normal operation */
  vs1001_hard_reset();

  return TRUE;
}

void vs1001_set_volume(u08_t left, u08_t right) {
  cbi(EIMSK, INT4);  // disable dreq irq
  /* now a dreq will not refill the buffer */
  
  /* wait for burst to end */
  while(burst_count);

  cbi(SPCR, SPIE);   // SPI interrupt disable
  
  vs1001_write(VS1001_VOL, (((u16_t)left) << 8) | (u16_t)right);
	
  // bug workaround
  if (left > right)             left = right;
  if (left > 11 && left < 24)   vs1001_write(VS1001_STATUS, 2);

  sbi(SPCR, SPIE);   // SPI interrupt enable
  sbi(EIMSK, INT4);  // enable dreq irq
}

